/* BEGIN LICENSE
 * Copyright © Blue Mind SAS, 2012-2016
 *
 * This file is part of BlueMind. BlueMind is a messaging and collaborative
 * solution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU Affero General Public License as
 * published by the Free Software Foundation (version 3 of the License).
 *
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See LICENSE.txt
 * END LICENSE
 */
package net.bluemind.directory.api;

import java.util.List;
import java.util.Set;

import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;

import net.bluemind.addressbook.api.VCard;
import net.bluemind.core.api.BMApi;
import net.bluemind.core.api.ListResult;
import net.bluemind.core.api.fault.ServerFault;
import net.bluemind.core.container.model.ContainerChangelog;
import net.bluemind.core.container.model.ContainerChangeset;
import net.bluemind.core.container.model.ItemValue;
import net.bluemind.core.task.api.TaskRef;

@BMApi(version = "3")
@Path("/directory/{domain}")
public interface IDirectory {

	/**
	 * Fetch root {@link DirEntry}
	 * 
	 * @return {@link DirEntry}
	 * @throws ServerFault
	 */
	@GET
	DirEntry getRoot() throws ServerFault;

	/**
	 * Fetch {@link DirEntry} from a path
	 * 
	 * @param path
	 *            path of entry
	 * 
	 * @return {@link DirEntry} entry
	 * @throws ServerFault
	 */
	@POST
	@Path("_entry")
	DirEntry getEntry(String path) throws ServerFault;

	/**
	 * Fetch {@link DirEntry}s from a path
	 * 
	 * @param path
	 *            path of entries
	 * 
	 * @return {@link DirEntry} child of path
	 * @throws ServerFault
	 */
	@POST
	@Path("_childs")
	List<DirEntry> getEntries(String path) throws ServerFault;

	/**
	 * Delete entry
	 * 
	 * @param path
	 * @throws ServerFault
	 */
	@DELETE
	@Path("{path}")
	public TaskRef delete(@PathParam(value = "path") String path) throws ServerFault;

	/**
	 * get entry vcard
	 * 
	 * @param path
	 * @throws ServerFault
	 */
	@GET
	@Path("{uid}/_vcard")
	public ItemValue<VCard> getVCard(@PathParam(value = "uid") String uid) throws ServerFault;

	/**
	 * Delete entry
	 * 
	 * @param path
	 * @throws ServerFault
	 */
	@DELETE
	@Path("_byentryuid/{entryUid}")
	public TaskRef deleteByEntryUid(@PathParam(value = "entryUid") String entryUid) throws ServerFault;

	/**
	 * ChangeLog of the container since
	 * 
	 * @param since
	 *            timestamp of first changes we want to retrieve
	 * @return {@link ContainerChangelog}
	 * @throws ServerFault
	 */
	@GET
	@Path("_changelog")
	public ContainerChangelog changelog(@QueryParam("since") Long since) throws ServerFault;

	/**
	 * ChangeSet of the container since
	 * 
	 * @param since
	 *            timestamp of first changes we want to retrieve
	 * 
	 * @throws ServerFault
	 */
	@GET
	@Path("_changeset")
	public ContainerChangeset<String> changeset(@QueryParam("since") Long since) throws ServerFault;

	@POST
	@Path("_search")
	public ListResult<ItemValue<DirEntry>> search(DirEntryQuery query) throws ServerFault;

	@GET
	@Path("entry-uid/{entryUid}")
	public DirEntry findByEntryUid(@PathParam("entryUid") String entryUid) throws ServerFault;

	@GET
	@Path("entry-uid/{entryUid}/icon")
	@Produces("image/png")
	public byte[] getEntryIcon(@PathParam("entryUid") String entryUid) throws ServerFault;

	@GET
	@Path("entry-uid/{entryUid}/photo")
	@Produces("image/png")
	public byte[] getEntryPhoto(@PathParam("entryUid") String entryUid) throws ServerFault;

	@GET
	@Path("_icon/{path}")
	@Produces("image/png")
	public byte[] getIcon(@PathParam("path") String path) throws ServerFault;

	@GET
	@Path("entry-uid/{entryUid}/rolesfor_")
	public Set<String> getRolesForDirEntry(@PathParam("entryUid") String entryUid) throws ServerFault;

	@GET
	@Path("ou-uid/{ouUid}/rolesfor_")
	public Set<String> getRolesForOrgUnit(@PathParam("ouUid") String orgUnitUid) throws ServerFault;

	@GET
	@Path("_byEmail/{email}")
	DirEntry getByEmail(@PathParam("email") String email);

}
